
#include "Socket.h"

#include <signal.h>
#ifndef SIGPIPE
#define SIGPIPE 13
#endif

int Socket::m_numSockets = 0;
#ifdef _WIN32
mutex_t Socket::m_mutex = CreateMutex(NULL, FALSE, NULL);
#else
mutex_t Socket::m_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

// Constructor, called by our child classes to set base class stuff.
Socket::Socket(int flags)
{
	m_flags	= flags;
	m_sock	= INVALID_SOCKET;
	
	Startup();

	// Avoid crashing if there is a SIGPIPE signal
	signal(SIGPIPE, SIG_IGN);
}

// Constructor, used when we accept a new connection and create a new Socket object
Socket::Socket(SOCKET sock, int flags)
{
	Startup();

	m_flags = flags;
	m_sock	= sock;	
}

// Destructor
Socket::~Socket()
{
	Close();
	Shutdown();
}

// Close the socket if it hasn't already been closed
void Socket::Close()
{
	// Don't bother if somebody called Close() directly
	if (m_sock != INVALID_SOCKET)
	{
#ifdef _WIN32
		if (closesocket(m_sock) == INVALID_SOCKET)
			cerr << "Socket::Close, closesocket failed" << endl;
#else
		if (close(m_sock) == INVALID_SOCKET)
			cerr << "Socket::Close, closesocket failed" << endl;
#endif
		m_sock = INVALID_SOCKET;
	}
}

// Send bytes to our communication peer.
//  buffer - char* buffer containing the data to send
//  len    - how many bytes to send 
// Returns the number actually sent, could be less than len.
// Returns -1 on an error.
int Socket::SendBytes(char* buffer, int len)
{
	int sent = send(m_sock, buffer, len, 0);
	if (sent == -1)
		cerr << "Socket::SendBytes, send failed" << endl;
	return sent;
}

// Recv bytes into the passed in buffer, up to buffer length of maxlen.
// Returns the number received.
// Returns -1 on an error.
int	Socket::RecvBytes(char* buffer, int maxlen)
{
	int recvd = recv(m_sock, buffer, maxlen, 0);
	if (recvd == -1)
		cerr << "Socket::RecvBytes, recv failed" << endl;
	return recvd;
}

// Handle any first-time socket startup code.
// Also increment counter of total Socket objects.
void Socket::Startup()
{
	// Mutual exclusion on our static counter
	MutexBlock mb(&m_mutex);

	// If we are currently zero, do the block, always increment
	if ((m_numSockets++) == 0)
	{
#ifdef _WIN32
		// For Windows, we need to fire up the Winsock DLL before we can do socket stuff.
		WORD wVersionRequested;
		WSADATA wsaData;
		int err;

		// Use the MAKEWORD(lowbyte, highbyte) macro declared in Windef.h 
		wVersionRequested = MAKEWORD(2, 2);

		err = WSAStartup(wVersionRequested, &wsaData);
		if (err != 0) 
		{
			// Tell the user that we could not find a usable Winsock DLL
			cerr << "Socket::Startup, WSAStartup failed" << endl;
			return;
		}
#endif
	}	
}

// Handle any last-time socket shutdown code.
// Also decrement the counter of total Socket objects.
void Socket::Shutdown()
{
	// Mutual exclusion on our static counter
	MutexBlock mb(&m_mutex);

	// If after subtracting we reach 0, then cleanup
	if ((--m_numSockets) == 0)
	{
#ifdef _WIN32
		// Windows specific socket shutdown code
		WSACleanup();
#endif
	}
}

// Return a structure containing information about our peer.
// Namely their IP address and port number.
PeerInfo Socket::GetPeerInfo()
{
	PeerInfo result;
	result.name[0] = '\0';
	result.port = -1;

	if (m_sock == INVALID_SOCKET)
		return result;

	socklen_t len;
	struct sockaddr_storage addr;
	char ipstr[INET6_ADDRSTRLEN];
	int port;

	len = sizeof(addr);
	getpeername(m_sock, (struct sockaddr*)&addr, &len);

	// Deal with both IPv4 and IPv6:
	if (addr.ss_family == AF_INET) 
	{
		struct sockaddr_in *s = (struct sockaddr_in *)&addr;
		port = ntohs(s->sin_port);
		inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr);
	} 
	else 
	{ 
		// AF_INET6
		struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
		port = ntohs(s->sin6_port);
		inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr);
	}
#ifdef _WIN32
	sprintf_s(result.name, "%s", ipstr);
#else
	sprintf(result.name, "%s", ipstr);
#endif
	result.port = port;

	return result;
}

// Set options on the socket based on our flags value
void Socket::SetOptions()
{
#ifdef _WIN32
	BOOL opt = TRUE;
#else
	int opt = 1;
#endif

	if (setsockopt(m_sock, SOL_SOCKET, SO_REUSEADDR, (char*) &opt, sizeof(opt)) == -1) 
		cerr << "Socket::SetOptions, setsockopt SO_REUSEADDR failed" << endl;	

	// TCP only options
	if (!(m_flags & DATAGRAM))
	{
		if (m_flags & NO_DELAY)
		{
			if (setsockopt(m_sock, SOL_SOCKET, TCP_NODELAY, (char*) &opt, sizeof(opt)) == -1) 
				cerr << "Socket::SetOptions, setsockopt TCP_NODELAY failed" << endl;
		}

		if (m_flags & KEEP_ALIVE)
		{
			if (setsockopt(m_sock, SOL_SOCKET, SO_KEEPALIVE, (char*) &opt, sizeof(opt)) == -1) 
				cerr << "Socket::SetOptions, setsockopt SO_KEEPALIVE failed" << endl;
		}
	}
}

