
#include "SocketServerDatagram.h"

SocketServerDatagram::SocketServerDatagram(int port, int flags, bool& good) : Socket(flags)
{
	int status;
	struct addrinfo hints;
	struct addrinfo *servinfo;			// will point to the results

	good = false;

	memset(m_lastPeer.name, 0, sizeof(m_lastPeer.name));
	m_lastPeer.port = 0;

	memset(&hints, 0, sizeof hints);	// make sure the struct is empty
	hints.ai_family   = AF_INET;        // specify IPv4 
	hints.ai_flags    = AI_PASSIVE;     // fill in my IP for me
	hints.ai_socktype = SOCK_DGRAM;		// UDP datagram service

	char portNum[16];
#ifdef _WIN32
	sprintf_s(portNum, "%d", port);
#else
	sprintf(portNum, "%d", port);
#endif

	status = getaddrinfo(NULL, portNum, &hints, &servinfo);	
	if (status != 0)
	{
	    cerr << "SocketServerDatagram::SocketServerDatagram, getaddrinfo failed" << endl;
	    return;
	}
	m_sock = socket(servinfo->ai_family, servinfo->ai_socktype, servinfo->ai_protocol);

	// Set some options on the socket
	SetOptions();

	if (bind(m_sock, servinfo->ai_addr, servinfo->ai_addrlen) == INVALID_SOCKET)
	{
		m_sock = INVALID_SOCKET;
		cerr << "SocketServerDatagram::SocketServerDatagram, bind failed" << endl;
		return;
	}

	freeaddrinfo(servinfo); // free the linked-list

	good = true;

}

// Send a datagram to the specified host and port
int	SocketServerDatagram::SendBytes(const char* host, int port, char* buffer, int maxlen)
{
	int status;
	struct addrinfo hints;
	struct addrinfo *servinfo;			// will point to the results

	memset(&hints, 0, sizeof hints);	// make sure the struct is empty
	hints.ai_family   = AF_INET;        // specify IPv4 
	hints.ai_flags    = AI_PASSIVE;     // fill in my IP for me
	hints.ai_socktype = SOCK_DGRAM;		// UDP datagram service

	char portNum[16];
#ifdef _WIN32
	sprintf_s(portNum, "%d", port);
#else
	sprintf(portNum, "%d", port);
#endif

	status = getaddrinfo(host, portNum, &hints, &servinfo);	
	if (status != 0)
	{
	    cerr << "SocketServerDatagram::SendBytes, getaddrinfo failed" << endl;
	    return -1;
	}

	int sent = sendto(m_sock, buffer, maxlen, 0, servinfo->ai_addr, servinfo->ai_addrlen);
	if (sent == -1)
		cerr << "SocketServerDatagram::SendBytes, send failed" << endl;

	freeaddrinfo(servinfo);

	return sent;
}

// Override the Socket::SendBytes method, we want to sendto() using last person who sent us a datagram
int	SocketServerDatagram::SendBytes(char* buffer, int maxlen)
{
	return SendBytes(m_lastPeer.name, m_lastPeer.port, buffer, maxlen);
}

// Override the Socket::RecvBytes method since we need to use recvfrom() for datagrams.
int	SocketServerDatagram::RecvBytes(char* buffer, int maxlen)
{
	struct sockaddr_storage their_addr;
	socklen_t addr_len = sizeof(their_addr);

	int recvd = recvfrom(m_sock, buffer, maxlen, 0, (struct sockaddr *) &their_addr, &addr_len);
	if (recvd == -1)
		cerr << "SocketServerDatagram::RecvBytes, recv failed" << endl;

	// Remember the info of the last sender
	if (their_addr.ss_family == AF_INET) 
	{
		struct sockaddr_in *s = (struct sockaddr_in *)&their_addr;
		m_lastPeer.port = ntohs(s->sin_port);
		inet_ntop(AF_INET, &s->sin_addr, m_lastPeer.name, sizeof(m_lastPeer.name));
	} 
	else 
	{ 
		// AF_INET6
		struct sockaddr_in6 *s = (struct sockaddr_in6 *)&their_addr;
		m_lastPeer.port = ntohs(s->sin6_port);
		inet_ntop(AF_INET6, &s->sin6_addr, m_lastPeer.name, sizeof(m_lastPeer.name));
	}

	return recvd;
}

// Find out info about the last peer we got a datagram from
PeerInfo SocketServerDatagram::GetPeerInfo()
{
	return m_lastPeer;
}



