// Test harness for the Socket class.  Can be run either as the server or the client.  
// Will echo a sequence of letters back and forth from client to server a specified number of times.

#include <iostream>
#include "SocketClient.h"
#include "SocketServer.h"
#include "SocketServerDatagram.h"
#include "HiResTimer.h"
#include "Utils.h"
#include "Threading.h"

// Fill a char* with cyclic sequences A-Z
void fillBuffer(char* buffer, int len)
{
	int i = 0;
	for (i = 0; i < len; i++)
		buffer[i] = (char) (i % 26 + 65);
}

using namespace std;

THREAD_START Worker(void *aParam)
{
	Socket* conn = (Socket*) aParam;
	double sendTime = 0.0;
	double RTTTime = 0.0;
	HiResTimer timer;
	HiResTimer timerRTT;
	unsigned long long sendBytes = 0;

	if (conn)
	{
		PeerInfo info = conn->GetPeerInfo();
		cout << "IP = " << info.name << ", port = " << info.port << endl;

		// Read the initial setup message from client.
		unsigned int setupMsg[4];
		int numRecv		= 0;
		int numSent		= 0;

		numRecv = conn->RecvBytes((char *) setupMsg, sizeof(unsigned int) * 4);
		unsigned int msgSize = setupMsg[0];
		unsigned int maxSend = setupMsg[1];
		unsigned int cycles  = setupMsg[2];
		bool oneWay = false;
		if (setupMsg[3] > 0)
			oneWay = true;

		cout << "SERVER: client wants msgSize = " << msgSize << ", maxSend = " << maxSend << ", cycles = " << cycles << ", oneway = " << oneWay << endl;
				
		// Send the client a single byte to let it know we got the setup message
		numSent= conn->SendBytes((char *) setupMsg, 1);
		cout << "SERVER: sent setup response of " << numSent << endl;

		// Make the buffer
		char* buffer = NULL;
		buffer = new char[msgSize];

		for (unsigned int i = 0; i < cycles; i++)
		{					
			// Receive as many times as it takes to get to the full message length
			cout << i << " recv ";
			unsigned int left = msgSize;
			char* current = buffer;
			unsigned int packets = 0;
			unsigned int bytes = 0;

			while (left > 0)
			{
				numRecv = conn->RecvBytes(buffer, msgSize);
				if (numRecv == -1)
				{
					cerr << "Failed to receive bytes!" << endl;
					delete buffer;
					buffer = NULL;
					return 0;
				}
				left -= numRecv;
				current += numRecv;
				packets++;
				bytes += numRecv;
			}
			cout << getTimeDateStamp(true, true) << " ";
			cout << packets << " packets, " << bytes << endl;
			
			if (!oneWay)
			{
				// Send as many times as it takes to get through the buffer
				left = msgSize;
				current = buffer;			
				cout << i << " send ";
				packets = 0;
				bytes = 0;
				timer.Reset();
				while (left > 0)
				{
					// Send at most maxSend bytes or up to the end of the buffer
					numSent = conn->SendBytes(buffer, min(maxSend, left));
					if (numSent == -1)
					{
						cerr << "Failed to send bytes!" << endl;
						delete buffer;
						buffer = NULL;
						return 0;
					}
					left -= numSent;
					current += numSent;
					packets++;
					bytes += numSent;

				}
				sendTime += timer.GetElapsed();
				sendBytes += bytes;
				
				cout << getTimeDateStamp(true, true) << " ";
				cout << packets << " packets, " << bytes << endl;
			}
		}

		printf("Send time             : %10.4f\n", sendTime);
		printf("Send bytes            : %10lld\n", sendBytes);
		printf("Send bandwidth (Mbps) : %10.4f\n", ((sendBytes * 8.0 / sendTime) / 1e6));
		
		delete buffer;
		buffer = NULL;

		delete conn;
		conn = NULL;
	}
	return 0;
}

int main(int argc, char** argv)
{		
	string	host		= "";
	int		port		= 0;

	// Default values for options that don't need to be set
	unsigned int	cycles		= 1;
	unsigned int	msgSize		= 1000;
	unsigned int	maxSend		= 1000;
	unsigned int	backlog		= 10;
	unsigned int    flags       = 0;
	unsigned int    delay       = 0;

	// Parse the command line switches
	if (!getSwitch(argc, argv, "port", &port))
	{
		cout << "TestSocket" << endl;
		cout << " -backlog   <server: maximum backlog>" << endl;
		cout << " -port      <client/server: port number>" << endl;
		cout << " -host      <client: hostname>" << endl;
		cout << " -msgsize   <client: message size in bytes>" << endl;
		cout << " -maxsend   <client: max bytes to send at one time>" << endl;
		cout << " -cycles    <client: send/recv cycles>" << endl;
		cout << " -printmsg  [dump the message buffer]" << endl;
		cout << " -threaded  [server: handle multiple connections]" << endl;
		cout << " -oneway    [client only sends, server only receives]" << endl;
		cout << " -flags     <bitwise OR of: 1=no delay, 2=keep alive, 4=datagram>" << endl;
		cout << " -delay     [client delay between sends, in ms]" << endl;

		cout << endl;

		return 0;
	}

	bool isClient = getSwitch(argc, argv, "host", &host);		
	bool printMsg = getSwitch(argc, argv, "printmsg");
	bool threaded = getSwitch(argc, argv, "threaded");
	bool oneWay   = getSwitch(argc, argv, "oneway");

	getSwitch(argc, argv, "cycles", &cycles);
	getSwitch(argc, argv, "maxsend", &maxSend);
	getSwitch(argc, argv, "msgsize", &msgSize);
	getSwitch(argc, argv, "backlog", &backlog);
	getSwitch(argc, argv, "flags", &flags);
	getSwitch(argc, argv, "delay", &delay);

	// Create some fake data to send
	char* buffer = new char[msgSize];
	fillBuffer(buffer, msgSize);

	HiResTimer timer;
	HiResTimer timerRTT;
	HiResTimer timerTotal;
	double sendTime = 0.0;
	double RTTTime = 0.0;

	unsigned long long sendBytes = 0;

	////////////////////////////////////////// client code TCP /////////////////////////////////////////
	if ((isClient) && !(flags & Socket::DATAGRAM))
	{
		cout << "Waiting to talk to " << host << " on port " << port << endl;
		timerTotal.Reset();

		bool good = false;
		SocketClient client(host.c_str(), port, flags, good);	
		int numSent = 0;
		int numRecv = 0;
			
		if (!good)
		{
			cerr << "Failed to create client connection" << endl;
			delete buffer;
			buffer = NULL;
			return 0;
		}

		PeerInfo info = client.GetPeerInfo();
		cout << "IP = " << info.name << ", port = " << info.port << endl;

		// Start by telling the server the client choices for msgSize, maxSend, and cycles.		
		// NOTE: this relies on both client and server being same Endianness.
		unsigned int setupMsg[4];
		setupMsg[0] = msgSize;
		setupMsg[1] = maxSend;
		setupMsg[2] = cycles;
		setupMsg[3] = 0;
		if (oneWay)
			setupMsg[3] = 1;

		numSent = client.SendBytes((char*) setupMsg, sizeof(unsigned int) * 4);
		cout << "CLIENT: setup message of " << numSent << endl;

		// Wait to hear a one-byte response from server
		numRecv = client.RecvBytes((char*) setupMsg, 1);
		cout << "CLIENT: got setup response of " << numRecv << endl;
		
		HiResTimer timerSendRecv;
		double totalSendRecv = 0.0;

		for (unsigned int i = 0; i < cycles; i++)
		{			
			// Send as many times as it takes to get through the buffer
			unsigned int left = msgSize;
			char* current = buffer;			
			cout << i << " send ";
			unsigned int bytes = 0;
			unsigned int packets = 0;
			timer.Reset();
			timerRTT.Reset();
			while (left > 0)
			{
				// Send at most maxSend bytes or up to the end of the buffer
				if (delay > 0)
					sleep_ms(delay);
				timerSendRecv.Reset();
				numSent = client.SendBytes(buffer, min(maxSend, left));
				if (numSent == -1)
				{
					cerr << "Failed to send bytes!" << endl;
					delete buffer;
					buffer = NULL;
					return 0;
				}				
				left -= numSent;
				current += numSent;
				packets++;
				bytes += numSent;

//cout << "packets = " << packets << endl;
			}
			sendTime += timer.GetElapsed();
			sendBytes += bytes;
			
			cout << getTimeDateStamp(true, true) << " ";
			cout << packets << " packets, " << bytes << endl;

			if (!oneWay)
			{
				// Receive as many times as it takes to get to the full message length
				cout << i << " recv ";
				left = msgSize;
				current = buffer;
				packets = 0;
				bytes = 0;
				while (left > 0)
				{
					numRecv = client.RecvBytes(buffer, msgSize);
					if (numRecv == -1)
					{
						cerr << "Failed to receive bytes!" << endl;
						delete buffer;
						buffer = NULL;
						return 0;
					}
					left -= numRecv;
					current += numRecv;
					packets++;
					bytes += numRecv;
				}
				RTTTime += timerRTT.GetElapsed();
				cout << getTimeDateStamp(true, true) << " ";
				cout << packets << " packets, " << bytes;

				// Optionally we print out the received message
				if (printMsg)
				{
					char ch = buffer[msgSize - 1];
					buffer[msgSize - 1] = '\0';
					cout << buffer << endl;
					buffer[msgSize - 1] = ch;
				}
			}
			double sendRecvTime = timerSendRecv.GetElapsed();
			cout << " send/recv time " << sendRecvTime << endl;
			totalSendRecv += sendRecvTime;
		}
		
		printf("Elapsed time          : %10.4f\n", timerTotal.GetElapsed());
		printf("Send time             : %10.4f\n", sendTime);
		printf("Send bytes            : %10lld\n", sendBytes);
		printf("Send bandwidth (Mbps) : %10.4f\n", ((sendBytes * 8.0 / sendTime) / 1e6));
		printf("RTT estimate          : %10.4f\n", RTTTime / cycles);
		printf("Send/recv avg time    : %10.4f\n", totalSendRecv / cycles);

	}	
	////////////////////////////////////////// client code UDP /////////////////////////////////////////
	else if ((isClient) && (flags & Socket::DATAGRAM))
	{
		cout << "Waiting to talk to " << host << " on port " << port << endl;
		timerTotal.Reset();

		bool good = false;
		SocketClient client(host.c_str(), port, flags, good);	
		int numSent = 0;
		int numRecv = 0;
			
		if (!good)
		{
			cerr << "Failed to create client connection" << endl;
			delete buffer;
			buffer = NULL;
			return 0;
		}

		PeerInfo info = client.GetPeerInfo();
		cout << "IP = " << info.name << ", port = " << info.port << endl;

		int MAX_LEN = 65536;
		char* datagram = new char[MAX_LEN];
		
		HiResTimer timerSendRecv;
		double totalSendRecv = 0.0;

		for (unsigned int i = 0; i < cycles; i++)
		{			
			// Send at most maxSend bytes or up to the end of the buffer
			if (delay > 0)
				sleep_ms(delay);
			timerSendRecv.Reset();

			
			
			cout << getTimeDateStamp(true, true) << " ";
			cout << i << ", ";
			numSent = client.SendBytes(buffer, msgSize);
			cout << "sent datagram, " << numSent << endl;

			if (!oneWay)
			{
				numRecv = client.RecvBytes(buffer, MAX_LEN);
				cout << getTimeDateStamp(true, true) << " ";
				cout << i << ", recv datagram, " << numRecv;
			}

			double sendRecvTime = timerSendRecv.GetElapsed();
			cout << " send/recv time " << sendRecvTime << endl;
			totalSendRecv += sendRecvTime;
		}
		
		printf("Elapsed time          : %10.4f\n", timerTotal.GetElapsed());
		printf("Send/recv avg time    : %10.4f\n", totalSendRecv / cycles);

		delete datagram;
		datagram = NULL;
	}

	////////////////////////////////////////// server code TCP /////////////////////////////////////////
	else if (!(flags & Socket::DATAGRAM))
	{
		bool good = false;
		SocketServer server(port, backlog, flags, good);

		if (!good)
		{	
			cerr << "Failed to create server connection" << endl;
			delete buffer;
			buffer = NULL;
			return 0;		
		}

		VECTOR_THREAD vecThreads;
		while (true)
		{
			// Shutdown the server if somebody hit Q (between recv/sends)
			unsigned char ch = getKeyAsync();
			if ((ch == 'q') || (ch == 'Q'))
			{
				// Wait for all clients to finish talking and free their resources
				join_all_threads(&vecThreads);
				delete_all_threads(&vecThreads);
				break;
			}

			Socket* conn = server.Accept();
			if (conn)
			{
				mythread_t hThread;
				create_thread_arg(&hThread, Worker, (void *) conn);

				// Single threaded mode, just wait for thread to finish
				if (!threaded)
				{
					cout << "WAITING for thread" << endl;
					join_thread(hThread);
					delete_thread(hThread);
				}
				else
				{
					vecThreads.push_back(hThread);
				}
			}
		}
	}
	////////////////////////////////////////// server code UDP /////////////////////////////////////////
	else if (flags & Socket::DATAGRAM)
	{
		bool good = false;
		SocketServerDatagram server(port, flags, good);

		if (!good)
		{	
			cerr << "Failed to create server connection" << endl;
			delete buffer;
			buffer = NULL;
			return 0;		
		}
		
		int i = 0;
		int MAX_LEN = 65536;
		char* datagram = new char[MAX_LEN];
		while (true)
		{
			// Shutdown the server if somebody hit Q (between recv/sends)
			unsigned char ch = getKeyAsync();
			if ((ch == 'q') || (ch == 'Q'))
				break;

			int numRecv = server.RecvBytes(datagram, MAX_LEN);
			if (numRecv != -1)
			{
				cout << getTimeDateStamp(true, true) << " ";
				cout << i << ", received datagram, " << numRecv;
			
				PeerInfo info = server.GetPeerInfo();
				cout << ", from " << info.name;
				cout << " port " << info.port << endl;

				if (!oneWay)
				{					
					int numSent = server.SendBytes(datagram, msgSize);
					cout << getTimeDateStamp(true, true) << " ";
					cout << i << ", sent datagram, " << numSent << endl;
				}
				i++;
			}			
		}
		delete datagram;
		datagram = NULL;
	}

	delete buffer;
	buffer = NULL;

	return 0;
}

